import numpy as np
from scipy.signal import welch

def delta_band(data):
    from .config import SAMPLING_FREQUENCY
    fs = SAMPLING_FREQUENCY

    delta_band = (0.5, 4) 
    delta_powers = []
    
    for channel_data in data:
        freqs, psd = welch(channel_data, fs=fs, nperseg=fs*2)
        
        delta_indices = np.logical_and(freqs >= delta_band[0], freqs <= delta_band[1])
        
        delta_power = np.trapz(psd[delta_indices], freqs[delta_indices])
        delta_powers.append(delta_power)
    
    return np.array(delta_powers)

def theta_band(data):
    from .config import SAMPLING_FREQUENCY
    fs = SAMPLING_FREQUENCY

    theta_band = (4, 7) 
    theta_powers = []
    
    for channel_data in data:
        freqs, psd = welch(channel_data, fs=fs, nperseg=fs*2)
        
        theta_indices = np.logical_and(freqs >= theta_band[0], freqs <= theta_band[1])
        
        theta_power = np.trapz(psd[theta_indices], freqs[theta_indices])
        theta_powers.append(theta_power)
    
    return np.array(theta_powers)

def alpha_band(data):
    from .config import SAMPLING_FREQUENCY
    fs = SAMPLING_FREQUENCY

    alpha_band = (8, 13) 
    alpha_powers = []
    
    for channel_data in data:
        freqs, psd = welch(channel_data, fs=fs, nperseg=fs*2)
        
        alpha_indices = np.logical_and(freqs >= alpha_band[0], freqs <= alpha_band[1])
        
        alpha_power = np.trapz(psd[alpha_indices], freqs[alpha_indices])
        alpha_powers.append(alpha_power)
    
    return np.array(alpha_powers)

def beta_band(data):
    from .config import SAMPLING_FREQUENCY
    fs = SAMPLING_FREQUENCY

    beta_band = (14, 30) 
    beta_powers = []
    
    for channel_data in data:
        freqs, psd = welch(channel_data, fs=fs, nperseg=fs*2)
        
        beta_indices = np.logical_and(freqs >= beta_band[0], freqs <= beta_band[1])
        
        beta_power = np.trapz(psd[beta_indices], freqs[beta_indices])
        beta_powers.append(beta_power)
    
    return np.array(beta_powers)

def beta_band_relative(data):
    from .config import SAMPLING_FREQUENCY
    fs = SAMPLING_FREQUENCY

    beta_band = (13, 30)     # Beta frequency range
    relevant_band = (4, 50)  # Total power range, excluding delta
    
    relative_beta_powers = []
    
    for channel_data in data:
        freqs, psd = welch(channel_data, fs=fs, nperseg=fs*2)
        
        beta_indices = np.logical_and(freqs >= beta_band[0], freqs <= beta_band[1])
        relevant_indices = np.logical_and(freqs >= relevant_band[0], freqs <= relevant_band[1])
        
        beta_power = np.trapz(psd[beta_indices], freqs[beta_indices])
        
        total_relevant_power = np.trapz(psd[relevant_indices], freqs[relevant_indices])
        
        relative_beta_power = beta_power / total_relevant_power if total_relevant_power > 0 else 0
        relative_beta_powers.append(relative_beta_power)
    
    return np.array(relative_beta_powers)

def alpha_band_relative(data):
    from .config import SAMPLING_FREQUENCY
    fs = SAMPLING_FREQUENCY

    beta_band = (8, 13)     # Beta frequency range
    relevant_band = (4, 50)  # Total power range, excluding delta
    
    relative_beta_powers = []
    
    for channel_data in data:
        freqs, psd = welch(channel_data, fs=fs, nperseg=fs*2)
        
        beta_indices = np.logical_and(freqs >= beta_band[0], freqs <= beta_band[1])
        relevant_indices = np.logical_and(freqs >= relevant_band[0], freqs <= relevant_band[1])
        
        beta_power = np.trapz(psd[beta_indices], freqs[beta_indices])
        
        total_relevant_power = np.trapz(psd[relevant_indices], freqs[relevant_indices])
        
        relative_beta_power = beta_power / total_relevant_power if total_relevant_power > 0 else 0
        relative_beta_powers.append(relative_beta_power)
    
    return np.array(relative_beta_powers)

def gamma_band(data):
    from .config import SAMPLING_FREQUENCY
    fs = SAMPLING_FREQUENCY

    gamma_band = (0.5, 4) 
    gamma_powers = []
    
    for channel_data in data:
        freqs, psd = welch(channel_data, fs=fs, nperseg=fs*2)
        
        gamma_indices = np.logical_and(freqs >= gamma_band[0], freqs <= gamma_band[1])
        
        gamma_power = np.trapz(psd[gamma_indices], freqs[gamma_indices])
        gamma_powers.append(gamma_power)
    
    return np.array(gamma_powers)


import numpy as np

def katz_fd(signal):
    katz_fds = []
    for channel in signal:
        N = len(signal)
        L = np.sum(np.abs(np.diff(channel)))
        d = np.max(np.abs(channel - channel[0]))
        katz = np.log(L) / np.log(d)
        katz_fds.append(katz)
    
    return katz_fds

def petrosian_fd(signal):
    petrosian_fds = []
    for channel in signal:
        N = len(channel)
        diff_signal = np.diff(channel)
        N_delta = np.sum(diff_signal[:-1] * diff_signal[1:] < 0)
        petrosian =  np.log(N) / (np.log(N) + np.log(N_delta))
        petrosian_fds.append(petrosian)
    
    return petrosian_fds

if __name__ == '__main__':
    for i in range(1000):
        signal = np.random.rand(10, 1000)
        print("Katz FD:", katz_fd(signal))
        print("Petrosian FD:", petrosian_fd(signal))



import pywt
import numpy as np


def get_wavelet_features(eeg_signal):
    from .config import SAMPLING_FREQUENCY, WAVELET_FREQUENCIES
    sampling_rate = SAMPLING_FREQUENCY
    wavelet_features = []
    for channel in eeg_signal:
        frequencies = WAVELET_FREQUENCIES
        scales = [sampling_rate / freq for freq in frequencies]
        coefficients, _ = pywt.cwt(channel, scales, 'cmor')

        # print(len(coefficients))

        mean_coeffs = np.mean(np.abs(coefficients), axis=1)
        std_coeffs = np.std(coefficients, axis=1)
        
        channel_wavelet_features = np.concatenate((mean_coeffs, std_coeffs), axis = 0)
        wavelet_features.append(channel_wavelet_features)

    return wavelet_features



import numpy as np
from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import ExtraTreesClassifier, RandomForestClassifier, GradientBoostingClassifier, AdaBoostClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.linear_model import LogisticRegression

from .time_features import *
from .frequency_features import *
from .entropy_features import *
from .wavelet_features import *
from .fractal_dimensions_features import *

#VISUALIZATION CONSTANTS

SPECTOGRAM_WINDOW_LENGTH = 20
SPECTOGRAM_STD_DEV = 12
SPECTOGRAM_HOP_SIZE = 2
SPECTOGRAM_MFFT = 800
SPECTROGRAM_MAX_FREQUENCY = 60
#MODEL CONSTANTS

SOLE_MODEL_ACCURACY_THRESHOLD = 0.5
SOLE_MODEL_TESTING = 0.2
SAMPLING_FREQUENCY = 250

GAF_CMAP = "binary_r"
#MODEL TRAINING

MODELS = {
    'svm': SVC(probability=True),
    'knn': KNeighborsClassifier(),
    'extra_trees': ExtraTreesClassifier(),
    'random_forest': RandomForestClassifier(),
    'decision_tree': DecisionTreeClassifier(),
    # 'gradient_boosting': GradientBoostingClassifier(),
    'ada_boost': AdaBoostClassifier(),
    'naive_bayes': GaussianNB(),
    # 'logistic_regression': LogisticRegression(),
}

PARAM_GRIDS = {
    'svm': {
        'C': np.logspace(-1, 10, 5),
        'kernel': ['rbf'],
    },
    'knn': {
        'n_neighbors': [1, 3, 5, 10, 20, 50, 100],
        'weights': ['uniform', 'distance'],
    },
    'extra_trees': {
        'n_estimators': [100, 200, 300, 500],
        'max_depth': [None, 10, 50, 100, 150],
        'min_samples_split': [2, 5, 10, 20], 
        'min_samples_leaf': [1, 2, 5, 10], 
    },
    'random_forest': {
        'n_estimators': [100, 200, 300, 500],
        'max_depth': [None, 10, 50, 100, 150],
        'min_samples_split': [2, 5, 10, 20],
        'min_samples_leaf': [1, 2, 5, 10], 
    },
    'decision_tree': {
        'max_depth': [None, 10, 50, 100, 150],
        'min_samples_split': [2, 5, 10, 20],
        'min_samples_leaf': [1, 2, 5, 10],
    },
    'gradient_boosting': {
        'n_estimators': [100, 200, 500],
        'learning_rate': [0.01, 0.1, 0.3],
        'max_depth': [3, 5, 7, 9],
    },
    'naive_bayes': {
        # GaussianNB has no hyperparameters to tune
    },
    'logistic_regression': {
        'C': np.logspace(-3, 3, 7),
        'penalty': ['l1', 'l2', 'elasticnet', 'none'],  
        'solver': ['saga', 'lbfgs', 'liblinear'],  
    },
}

#MODEL TRAINING CONSTANTS

SEARCH_VERBOSE = 1
SCORING_METHOD = "accuracy"
SEARCH_ITERATIONS = 40
CROSS_VALIDATION_NUM = 4


#FEATURES

TIME_FEATURES_ARRAY = [
    '01_std',
    '01_ptp',
    '01_var',
    '01_min',
    '01_max',
    '01_argmin',
    '01_argmax',
    '01_rms',
    '01_abs_diff_signal',
    '01_skewness',
    '01_kurtosis',
    '01_signal_energy',
    '01_mean',
    '01_hjorth_complexity',
    '01_hjorth_mobility',
]

FREQUENCY_FEATURES_ARRAY = [
    '01_theta_channels',
    '01_alpha_channels',
    '01_beta_channels',
    # '01_gamma_channels',
    # '01_delta_channels'
]

# ENTROPY_FEATURES_ARRAY = [
#     'wavelet_entropy',
#     'spectral_entropy_channels'
#     'sample_entropy',
#     'wavelet_entropy',
#     'fuzzy_entropy',
#     'spectral_fuzzy_entropy'
# ]

FRACTAL_DIMENSIONS_FEATURES_ARRAY = [
    '01_katz',
    '01_petrosian',
]

WAVELET_FEATURES_ARRAY = [
    '24_mean_and_std_wavelet_coefficients'
]

ALL_FEATURES_ARRAY = TIME_FEATURES_ARRAY + FREQUENCY_FEATURES_ARRAY + WAVELET_FEATURES_ARRAY + FRACTAL_DIMENSIONS_FEATURES_ARRAY

FEATURE_FUNCTIONS_DICT = {

    # Time Based Features
    '01_std': std,
    '01_ptp': ptp,
    '01_var': var,
    '01_min': min,
    '01_max': max,
    '01_argmin': argmin,
    '01_argmax': argmax,
    '01_rms': rms,
    '01_abs_diff_signal': abs_diff_signal,
    '01_skewness': skewness,
    '01_kurtosis': kurtosis,
    '01_signal_energy': signal_energy,
    '01_mean': mean,
    '01_hjorth_complexity': hjorth_complexity,
    '01_hjorth_mobility': hjorth_mobility,

    #Frequnecy Based Features
    '01_theta_channels': theta_band,
    '01_alpha_channels': alpha_band,
    '01_beta_channels' : beta_band,
    # 'delta_channels': delta_band,

    #Fractal Dimension Features

    '01_katz' : katz_fd,
    '01_petrosian' : petrosian_fd,

    #Wavelet Features

    '24_mean_and_std_wavelet_coefficients' : get_wavelet_features

    }

#FEATURE CONSTANTS

WAVELET_FREQUENCIES = [61.115, 43.59, 31.09, 22.17, 15.81, 11.28, 8.04, 5.74, 4.09, 2.92, 2.08, 1.48]


import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score, roc_auc_score, confusion_matrix, classification_report
import matplotlib.pyplot as plt
from scipy.signal import spectrogram
import seaborn as sns
from tqdm import tqdm
from .utils import *
from .visualization import *

def ensure_list(value):
    """Ensure that the input value is wrapped in a list if it's not already a list."""
    if not isinstance(value, list):
        return [value]
    return value


def compute_spectrogram(eeg_signal, fs, nperseg=256, overlap_fraction=0.5, cmap='viridis'):
    """
    Calculate and plot the spectrogram of an EEG signal with adaptive overlap.

    Parameters:
    - eeg_signal: 1D numpy array of EEG signal data
    - fs: Sampling frequency of the EEG signal (default is 1000 Hz)
    - nperseg: Length of each segment for STFT (default is 256)
    - overlap_fraction: Fraction of `nperseg` to use for overlap (default is 0.5)
    - cmap: Colormap for the spectrogram plot (default is 'viridis')

    Returns:
    - None (plots the spectrogram)
    """
    noverlap = int(nperseg * overlap_fraction)
    
    frequencies, times, Sxx = spectrogram(eeg_signal, fs=fs, nperseg=nperseg, noverlap=noverlap)

    plt.figure(figsize=(10, 6))
    plt.pcolormesh(times, frequencies, 10 * np.log10(Sxx), shading='gouraud', cmap=cmap)
    plt.colorbar(label='Power/Frequency (dB/Hz)')
    plt.title("Spectrogram of EEG Signal")
    plt.xlabel("Time (s)")
    plt.ylabel("Frequency (Hz)")
    plt.ylim(0, 60)

    plt.tight_layout()
    plt.show()

def feature_selection(feature_array, label_array, all_features):
    from featurewiz import featurewiz
    from statistics import mode

    df = pd.DataFrame(feature_array, columns=all_features)
    df['diagnosis'] = label_array
    output = featurewiz(df, target='diagnosis', skip_sulov = True, corr_limit = 2)

    array1 = np.array(output[0])
    array2 = np.array(output[1])
    remove_array = []
    for data in array2:
        temp = []
        for i, number in enumerate(data):
            if number == 1 or number == 0:
                remove_array.append(i)

    remove = mode(remove_array)
    feature_array = []
    for data in array2:
        temp = []
        for i, number in enumerate(data):
            if i == remove:
                pass
            else:
                temp.append(number)
        feature_array.append(temp)
    feature_array = np.array(feature_array)
    return feature_array, array1

def evaluate_model_performance(classifier, X_test, y_test):
    """
    Evaluate the performance of a classifier with various metrics.

    Args:
    - classifier: The trained model (e.g., classifier)
    - X_test: Features of the test set
    - y_test: True labels for the test set

    Returns:
    - None: Prints various performance metrics to the console.
    """
    #Make your own predict function
    preds = classifier.predict(X_test)

    print(f"New Feature Array Shape: {X_test.shape}")
    
    print(f'Preds Length: {len(preds)}')

    final_accuracy = accuracy_score(y_test, preds)
    print(f'Final Accuracy: {final_accuracy}')
    
    f1 = f1_score(y_test, preds)
    print(f'F1 Score: {f1}')
    
    precision = precision_score(y_test, preds)
    print(f'Precision: {precision}')
    
    recall = recall_score(y_test, preds)
    print(f'Recall: {recall}')
    
    if hasattr(classifier, "predict_proba"):
        probs = classifier.predict_proba(X_test)[:, 1]
        auc_score = roc_auc_score(y_test, probs)
        print(f'ROC AUC: {auc_score}')
    
    conf_matrix = confusion_matrix(y_test, preds)
    print(f'Confusion Matrix:\n{conf_matrix}')
    
    class_report = classification_report(y_test, preds)
    print(f'Classification Report:\n{class_report}')


def apply_cmap(arr, cmap_name='viridis'):
    """
    Converts a 2D array into an RGB image using the specified colormap.

    Parameters:
    - arr (np.array): A 2D array with shape (256, 256), values in any range.
    - cmap_name (str): The name of the colormap to use (e.g., 'viridis', 'plasma', 'inferno', etc.).

    Returns:
    - rgb_image (np.array): A 3D RGB image with shape (256, 256, 3).
    """
    arr_normalized = (arr - np.min(arr)) / (np.max(arr) - np.min(arr))

    cmap = plt.get_cmap(cmap_name)

    rgba_image = cmap(arr_normalized)

    rgb_image = rgba_image[..., :3]

    return rgb_image

def display_image(image):
    """
    Function to display an image array.
    
    Parameters:
    image (numpy.ndarray): A 3D array of shape (X, X, 3) representing the image.
    """
    plt.imshow(image)
    plt.axis('off')
    plt.show()

def graph_training_history(history):
    plt.figure(figsize=(12, 6))

    plt.subplot(1, 2, 1)
    plt.plot(history.history['accuracy'], label='Training Accuracy')
    plt.plot(history.history['val_accuracy'], label='Validation Accuracy')
    plt.title('Model Accuracy')
    plt.xlabel('Epoch')
    plt.ylabel('Accuracy')
    plt.legend(loc='upper left')

    plt.subplot(1, 2, 2)
    plt.plot(history.history['loss'], label='Training Loss')
    plt.plot(history.history['val_loss'], label='Validation Loss')
    plt.title('Model Loss')
    plt.xlabel('Epoch')
    plt.ylabel('Loss')
    plt.legend(loc='upper right')

    plt.tight_layout()
    plt.show()

def display_image_classifier_performance_metrics(results_dict):
    """
    Prints and plots the performance metrics from the results dictionary.
    
    Parameters:
    - results_dict (dict): Dictionary containing the model's performance metrics.
    """
    # Print the metrics
    print(f"Test Accuracy: {results_dict['test_accuracy']:.4f}")
    print(f"Precision: {results_dict['precision']:.4f}")
    print(f"Recall: {results_dict['recall']:.4f}")
    print(f"F1 Score: {results_dict['f1_score']:.4f}")
    
    # Confusion Matrix
    cm = results_dict['confusion_matrix']
    print("Confusion Matrix:")
    print(cm)
    
    # Plot confusion matrix
    plt.figure(figsize=(8, 6))
    sns.heatmap(cm, annot=True, fmt='d', cmap='Blues', xticklabels=np.arange(cm.shape[0]), yticklabels=np.arange(cm.shape[0]))
    plt.title('Confusion Matrix')
    plt.xlabel('Predicted')
    plt.ylabel('True')
    plt.show()

    # Plot training history (Accuracy and Loss)
    history = results_dict['history']
    
    # Plot Accuracy
    plt.figure(figsize=(12, 6))
    plt.subplot(1, 2, 1)
    plt.plot(history['accuracy'], label='Training Accuracy')
    plt.plot(history['val_accuracy'], label='Validation Accuracy')
    plt.title('Accuracy over Epochs')
    plt.xlabel('Epochs')
    plt.ylabel('Accuracy')
    plt.legend()
    
    # Plot Loss
    plt.subplot(1, 2, 2)
    plt.plot(history['loss'], label='Training Loss')
    plt.plot(history['val_loss'], label='Validation Loss')
    plt.title('Loss over Epochs')
    plt.xlabel('Epochs')
    plt.ylabel('Loss')
    plt.legend()
    
    plt.tight_layout()
    plt.show()

def get_channel_feature_names(channel_names, all_features):
    channel_feature_names = []
    for channel in channel_names:
        for feature in all_features:
            number = int(feature[0] + feature[1])
            for num in range(number):
                channel_feature_names.append(f"{channel}_{feature}_{num}")
    return channel_feature_names

def get_gaf_image_array(data, max, min, gaf_scale, gaf_image_preview = True):
    images_array = []
    for epoch in tqdm(data):
        image = []
        for channel in epoch:
            one_channel_gaf = compute_gaf(channel, max, min, scale = gaf_scale)
            
            if gaf_image_preview:
                display_image(apply_cmap(one_channel_gaf, GAF_CMAP))
                gaf_image_preview = False 

            image.append(one_channel_gaf)
        images_array.append(image)
    images_array = np.array(images_array)
    A, B, C, D = images_array.shape 
    return images_array.reshape((A, C, D, B)) 



from .config import *
from .utils import *
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV, train_test_split
from sklearn.metrics import accuracy_score
from sklearn.ensemble import VotingClassifier
from sklearn.utils import all_estimators
import tensorflow as tf
from tensorflow.keras import layers, models
from tensorflow.keras.applications import ResNet50, VGG16
from tensorflow.keras.preprocessing import image
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.applications.resnet50 import preprocess_input
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras import layers, models, regularizers
from tensorflow.keras.callbacks import ModelCheckpoint
from sklearn.preprocessing import LabelEncoder
from pytorch_tabnet.tab_model import TabNetClassifier
import torch
from tabpfn import TabPFNClassifier
from sklearn.utils import resample


def create_and_fit_voting_classifier(models, param_grids, X_train, y_train, X_test, voting='hard'):
    """
    Create and fit a Voting Classifier based on given models and parameters.

    Parameters:
    - models (list): List of model instances (e.g., classifiers like LogisticRegression, SVM, etc.).
    - param_grids (list): List of dictionaries containing hyperparameter grids for each model.
    - X (array-like): Feature data.
    - y (array-like): Label data.
    - voting (str): 'hard' for majority voting, 'soft' for probability averaging.

    Returns:
    - fitted_voting_classifier (VotingClassifier): The trained VotingClassifier.
    """

    tuned_models = []
    
    for model, param_grid in zip(models, param_grids):
        param_grid = {key: ensure_list(value) for key, value in param_grid.items()}
        
        grid_search = GridSearchCV(model, param_grid, cv=5)
        tuned_models.append(('best_' + model.__class__.__name__, grid_search))
    
    voting_classifier = VotingClassifier(estimators=tuned_models, voting=voting)
    
    voting_classifier.fit(X_train, y_train)
    X_train, X_val, y_train, y_val = train_test_split(X_train, y_train, test_size=0.2, random_state=42)
    val_preds = voting_classifier.predict(X_val)
    val_accuracy = accuracy_score(y_val, val_preds)
    print(f"Validation Accuracy : {val_accuracy}")
    
    test_pred = voting_classifier.predict(X_test)
    return voting_classifier, val_accuracy, test_pred

def sole_models_hp_tuning(features_array, label_array, test_size, accuracy_threshold=0.9):
    """
    Perform conditional hyperparameter tuning for multiple models.
    If a model's baseline accuracy is below the threshold, skip tuning.
    """
    X_train, X_test, y_train, y_test = train_test_split(features_array, label_array, test_size=test_size)
    models = MODELS.copy()
    param_grids = PARAM_GRIDS.copy()

    accuracy_scores = {}
    best_params = {}
    best_model_info = {"models": []}
    max_accuracy = 0

    for model_name, model in models.items():
        # Step 1: Evaluate baseline accuracy with default parameters
        model.fit(X_train, y_train)
        baseline_pred = model.predict(X_test)
        baseline_accuracy = accuracy_score(y_test, baseline_pred)

        print(f"{model_name} baseline accuracy: {baseline_accuracy:.4f}")

        # Step 2: Skip hyperparameter tuning if baseline accuracy is below the threshold
        if baseline_accuracy < accuracy_threshold:
            print(f"Skipping hyperparameter tuning for {model_name} (baseline below threshold)")
            accuracy_scores[model_name] = baseline_accuracy
            continue

        # Step 3: Perform hyperparameter tuning
        param_grid = param_grids.get(model_name, {})
        grid_search = RandomizedSearchCV(
            estimator=model,
            param_distributions=param_grid,
            cv=CROSS_VALIDATION_NUM,
            scoring=SCORING_METHOD,
            n_iter=SEARCH_ITERATIONS,
            verbose=SEARCH_VERBOSE,
            n_jobs=-1,
        )
        grid_search.fit(X_train, y_train)

        # Best model and accuracy after tuning
        best_model = grid_search.best_estimator_
        y_pred = best_model.predict(X_test)
        accuracy = accuracy_score(y_test, y_pred)

        accuracy_scores[model_name] = accuracy
        best_params[model_name] = grid_search.best_params_

        # Step 4: Add to results if accuracy meets the threshold
        if accuracy >= accuracy_threshold:
            best_model_info["models"].append({
                "model_name": model_name,
                "model": best_model,
                "best_params": grid_search.best_params_,
                "accuracy": accuracy,
            })
            print(f"{model_name} tuned accuracy: {accuracy:.4f}")
            max_accuracy = max([max_accuracy, accuracy])

    return {"best_model_info": best_model_info, "accuracy_scores": accuracy_scores}, max_accuracy



def train_and_evaluate_image_classifier(data, test_data_real, labels, num_classes, epochs=10, batch_size=32):
    """
    Splits the data into training and testing sets, builds, compiles, and trains a CNN model.
    Evaluates the model on the test set using the best validation accuracy achieved during training.
    
    Parameters:
    - data (numpy.ndarray): Input data with shape (samples, 51, 51, 63)
    - labels (numpy.ndarray): Labels with shape (samples,)
    - num_classes (int): Number of classes for classification
    - epochs (int): Number of epochs to train the model
    - batch_size (int): Batch size for training
    
    Returns:
    - model: Trained TensorFlow model with the best weights
    - history: Training history object
    - test_accuracy: Accuracy of the model on the test set
    
    """
    label_encoder = LabelEncoder()
    labels = label_encoder.fit_transform(labels)
    labels = to_categorical(labels, num_classes=num_classes)
    
    train_data, test_data, train_labels, test_labels = train_test_split(data, labels, test_size=0.2, random_state=42)
    
    def create_model(input_shape, num_classes):
        """
        Creates a CNN model for classification tasks with input shape.
        
        Parameters:
        - input_shape (tuple): Shape of the input data (height, width, channels).
        - num_classes (int): Number of output classes.
        
        Returns:
        - model (tf.keras.Model): Compiled CNN model.
        """
        model = models.Sequential()
        
        model.add(layers.Conv2D(32, (3, 3), activation='relu', padding='same',
                                kernel_regularizer=regularizers.l2(0.02), input_shape=input_shape))

        model.add(layers.BatchNormalization())
        model.add(layers.MaxPooling2D((2, 2)))
        model.add(layers.Dropout(0.3))
        
        model.add(layers.Conv2D(64, (3, 3), activation='relu', padding='same',
                                kernel_regularizer=regularizers.l2(0.01)))
        model.add(layers.BatchNormalization())
        model.add(layers.MaxPooling2D((2, 2)))
        model.add(layers.Dropout(0.2))
        
        model.add(layers.Conv2D(128, (3, 3), activation='relu', padding='same',
                                kernel_regularizer=regularizers.l2(0.05)))
        model.add(layers.BatchNormalization())
        model.add(layers.MaxPooling2D((2, 2)))
        model.add(layers.Dropout(0.2))
        
        model.add(layers.Flatten())
        model.add(layers.Dense(256, activation='relu', kernel_regularizer=regularizers.l2(0.02)))
        model.add(layers.Dropout(0.2))
        
        model.add(layers.Dense(num_classes, activation='softmax'))
        
        model.compile(optimizer='adam',
                      loss='categorical_crossentropy',
                      metrics=['accuracy'])
        
        return model

    input_shape = data.shape[1:]
    model = create_model(input_shape=input_shape, num_classes=num_classes)
    best_val_accuracy = 0.0
    
    def update_best_val_accuracy(epoch, logs):
        nonlocal best_val_accuracy
        val_accuracy = logs.get('val_accuracy')
        if val_accuracy > best_val_accuracy:
            best_val_accuracy = val_accuracy

    checkpoint = ModelCheckpoint(
        filepath='best_model.keras',       
        monitor='val_accuracy',           
        save_best_only=True,               
        mode='max',                        
        verbose=1                       
    )
    best_accuracy_callback = tf.keras.callbacks.LambdaCallback(
        on_epoch_end=lambda epoch, logs: update_best_val_accuracy(epoch, logs)
    )
    history = model.fit(train_data, train_labels, epochs=epochs, batch_size=batch_size,
                        validation_split=0.2, callbacks=[checkpoint, best_accuracy_callback])

    model.load_weights('best_model.keras')
    
    test_loss, test_accuracy = model.evaluate(test_data, test_labels, verbose=0)
    
    predictions = model.predict(test_data)
    predicted_labels = np.argmax(predictions, axis=1)
    true_labels = np.argmax(test_labels, axis=1)
    
    precision = precision_score(true_labels, predicted_labels, average='weighted', zero_division=1)
    recall = recall_score(true_labels, predicted_labels, average='weighted', zero_division=1)
    f1 = f1_score(true_labels, predicted_labels, average='weighted', zero_division=1)
    
    cm = confusion_matrix(true_labels, predicted_labels)
    
    results_dict = {
        'test_accuracy': test_accuracy,
        'precision': precision,
        'recall': recall,
        'f1_score': f1,
        'confusion_matrix': cm,
        'history': history.history
    }
    test_pred = model.predict(test_data_real)

    return model, results_dict, best_val_accuracy, test_pred



def train_and_test_tabnet(X_train, y_train, X_test, epochs):
    
    model = TabNetClassifier()
    X_train, X_val, y_train, y_val = train_test_split(X_train, y_train, test_size=0.2, random_state=42)

    model.fit(X_train, y_train, max_epochs=epochs, patience=10, batch_size=128, virtual_batch_size=128)
    
    y_pred_val = model.predict(X_val)
    
    val_accuracy = accuracy_score(y_val, y_pred_val)

    print(f"Validation Accuracy: {val_accuracy}")

    test_pred = model.predict(X_test)
    
    return model, val_accuracy, test_pred


def preprocess_features_for_tabpfn(feature_array, label_array, max_samples=1024, max_features=99):
    if len(feature_array) > max_samples:
        feature_array, label_array = resample(
            feature_array, label_array, n_samples=max_samples, stratify=label_array, random_state=42
        )

    return feature_array, label_array

def train_tabpfn_classifier(X_train, y_train, X_test):
    X_train, y_train = preprocess_features_for_tabpfn(X_train[:, :99], y_train)

    X_train, X_val, y_train, y_val = train_test_split(X_train, y_train, test_size=0.2, random_state=42)

    classifier = TabPFNClassifier(device='cpu', N_ensemble_configurations=32)
    classifier.fit(X_train, y_train)

    val_pred, p_eval = classifier.predict(X_val, return_winning_probability=True)
    val_accuracy = accuracy_score(y_val, val_pred)
    print('Validation Accuracy:', val_accuracy)

    test_pred = classifier.predict(X_test[:, :99])

    return classifier, val_accuracy, test_pred



print("Importing Modules Please wait...")
import numpy as np
import pandas as pd
from .config import *
from .feature_extraction import extract_features, estimate_feature_extraction_time, extract_single_epoch_features
from tqdm import tqdm
from sklearn.model_selection import train_test_split
from .utils import *
from .model_training import *
from .visualization import compute_gaf, compute_spectrogram, plot_gaf, plot_spectrogram



class EEGPipeline:
    def __init__(self, data_array, label_array, sampling_rate, channel_names = None):
        self.data_array = data_array
        self.label_array = label_array
        self.channel_names = channel_names
        self.sf = sampling_rate
        self.num_classes = len(set(self.label_array))
        self.classifier = None
        self.feature_extractor = None
        SAMPLING_FREQUENCY = sampling_rate
        if channel_names is None:
            self.channel_names = ["channel_" + str(i) for i in range(data_array.shape[1])]
        if len(data_array.shape) != 3:
            raise ValueError(f"Data array must have 3 dimensions, but got {len(data_array.shape)}")
        if data_array.shape[0] != len(label_array):
            raise ValueError(f"Number of samples in data array does not match number of labels: {data_array.shape[0]} != {len(label_array)}")
        
        if len(self.channel_names) != data_array.shape[1]:
            raise ValueError(f"Number of channels in channel_names_array does not match the second dimension of data_array: {len(channel_names)} != {data_array.shape[1]}")
        
        # times_dict = estimate_feature_extraction_time(self.data_array)
        print("EEG Pipeline Initialized:")
        print(f"  Data shape: {data_array.shape} (Epochs, Channels, Timepoints)")
        print(f"  Number of Epochs and Labels: {len(label_array)}")
        print(f"  Number of classes: {self.num_classes}")
        print(f"  Number of channels: {len(self.channel_names)}")
        print(f"  Sampling rate: {sampling_rate} Hz")
        # encoder = LabelEncoder()
        # self.label_array = encoder.fit_transform(self.label_array)

        # print("Estimated Feature Extraction Times:")
        # print(f"  Time Domain: {times_dict['time']} seconds")
        # print(f"  Frequency Domain: {times_dict['frequency']} seconds")
        # print(f"  Entropy Domain: {times_dict['entropy']} seconds")
        # print(f"  Combined Domains: {times_dict['all']} seconds\n")

    class FeatureExtractor:
        def __init__(self, channel_feature_names, all_feature_names, filtered_features):
            self.channel_feature_names = channel_feature_names
            self.all_feature_names = all_feature_names
            self.filtered_features = filtered_features

        def extract_features(self, data):
            all_feature_array = np.array([extract_single_epoch_features(data, self.all_feature_names)])
            all_feature_df = pd.DataFrame(all_feature_array, columns = self.channel_feature_names)
            filtered_feature_df = all_feature_df[self.filtered_features]
            filtered_features = filtered_feature_df.to_numpy()
            filtered_features = np.nan_to_num(filtered_features, nan = 0)
            return filtered_features
        
    def train_model(
        self,
        initial_test_size: float = 0.2,
        enable_feature_selection: bool = False,
        time_features: bool = True,
        frequency_features: bool = False,
        # entropy_features: bool = False,
        wavelet_features: bool = False,
        fractal_dimensions_features: bool = False,
        voting_type: str = "hard",
        ml_baseline_accuracy: float = 0.8,
        tabnet_training_epochs: int = 200,
        gaf_training_epochs: int = 100,
        gaf_image_preview: bool = True,
        gaf_training_batch: int = 32,
        selected_classifiers: list = None, 
    ) -> None:
        """
        A configurable pipeline for feature extraction, model training, and prediction.

        Parameters:
        - initial_test_size: Test split ratio.
        - enable_feature_selection: Whether to perform feature selection.
        - time_features, frequency_features, etc.: Toggle different feature types.
        - voting_type: Voting method for ensemble classifiers.
        - enable_ensemble_machine_learning: Use ensemble ML if enabled.
        - tabnet_training_epochs, gaf_training_epochs: Training epochs for respective models.
        - selected_classifiers: List of classifiers to include ['gaf', 'tabpfn', 'tabnet', 'ml_ensemble'].
        """
        if selected_classifiers is None:
            selected_classifiers = ["gaf", "tabpfn", "tabnet", "ml_ensemble"]
        
        SOLE_MODEL_ACCURACY_THRESHOLD = ml_baseline_accuracy
        
        # Step 1: Train-test split
        X_train, X_test, y_train, y_test = train_test_split(
            self.data_array, self.label_array, test_size=initial_test_size, random_state=42
        )
        self.X_train, self.X_test, self.y_train, self.y_test = X_train, X_test, y_train, y_test

        # Step 2: Feature extraction
        all_feature_names = []
        if time_features:
            all_feature_names.extend(TIME_FEATURES_ARRAY)
        if frequency_features:
            all_feature_names.extend(FREQUENCY_FEATURES_ARRAY)
        if fractal_dimensions_features:
            all_feature_names.extend(FRACTAL_DIMENSIONS_FEATURES_ARRAY)
        if wavelet_features:
            all_feature_names.extend(WAVELET_FEATURES_ARRAY)

        self.all_feature_names = all_feature_names
        channel_feature_names = get_channel_feature_names(self.channel_names, all_feature_names)
        self.channel_feature_names = channel_feature_names

        feature_array = extract_features(X_train, all_feature_names)
        features_df = pd.DataFrame(feature_array, columns=channel_feature_names)

        if enable_feature_selection:
            filtered_feature_array, filtered_features = feature_selection(
                features_df.to_numpy(), y_train, channel_feature_names
            )
        else:
            filtered_feature_array = feature_array
            filtered_features = channel_feature_names

        self.filtered_features = filtered_features
        filtered_feature_array = np.nan_to_num(filtered_feature_array, nan=0)
        test_feature_array = self.get_test_features_array()


        # Step 3: Train classifiers
        classifiers = {}

        if "gaf" in selected_classifiers:
            gaf_classifier, gaf_val_accuracy, gaf_preds = self.train_model_gaf(
                gaf_training_epochs, gaf_image_preview, gaf_training_batch
            )
            classifiers["gaf_classifier"] = {
                "Classifier": gaf_classifier,
                "Val Accuracy": gaf_val_accuracy,
                "Predictions": gaf_preds,
            }

        if "tabpfn" in selected_classifiers:
            tabpfn_classifier, tabpfn_val_accuracy, tabpfn_preds = train_tabpfn_classifier(
                filtered_feature_array, y_train, test_feature_array
            )
            classifiers["tabpfn_classifier"] = {
                "Classifier": tabpfn_classifier,
                "Val Accuracy": tabpfn_val_accuracy,
                "Predictions": tabpfn_preds,
            }

        if "tabnet" in selected_classifiers:
            tabnet_classifier, tabnet_val_accuracy, tabnet_preds = train_and_test_tabnet(
                filtered_feature_array, y_train, test_feature_array, tabnet_training_epochs
            )
            classifiers["tabnet_classifier"] = {
                "Classifier": tabnet_classifier,
                "Val Accuracy": tabnet_val_accuracy,
                "Predictions": tabnet_preds,
            }

        if "ml_ensemble" in selected_classifiers:
            sole_model_info, _ = sole_models_hp_tuning(
                filtered_feature_array, y_train, SOLE_MODEL_TESTING, SOLE_MODEL_ACCURACY_THRESHOLD
            )
            best_models = [
                model_info["model"] for model_info in sole_model_info["best_model_info"]["models"]
            ]
            best_model_params = [
                model_info["best_params"] for model_info in sole_model_info["best_model_info"]["models"]
            ]
            ml_ensemble_classifier, ml_ensemble_val_accuracy, ml_ensemble_preds = create_and_fit_voting_classifier(
                best_models, best_model_params, filtered_feature_array, y_train, test_feature_array, voting=voting_type
            )
            classifiers["ml_ensemble_classifier"] = {
                "Classifier": ml_ensemble_classifier,
                "Val Accuracy": ml_ensemble_val_accuracy,
                "Predictions": ml_ensemble_preds,
            }

        best_classifier = None
        best_classifier_name = None
        best_test_accuracy = 0
        for clf_name, clf_info in classifiers.items():
            val_acc = clf_info["Val Accuracy"]
            test_pred = clf_info["Predictions"]
            test_acc = accuracy_score(y_test, test_pred)
            print(f"{clf_name} Validation Accuracy: {val_acc}")
            print(f"{clf_name} Testing Accuracy: {test_acc}")
            if test_acc > best_test_accuracy:
                best_test_accuracy = test_acc
                best_classifier = clf_info["Classifier"]
                best_classifier_name = clf_name

        self.classifiers = classifiers
        self.best_classifier = best_classifier
        print(f"Final Accuracy: {best_test_accuracy} achieved by {best_classifier_name}")
        self.feature_extractor = self.FeatureExtractor(self.channel_feature_names, self.all_feature_names, self.filtered_features)

        
    def extract_features_from_data(self, data_array):
        extracted_features = extract_features(data_array, self.all_feature_names)
        test_features_df = pd.DataFrame(extract_features, columns = self.channel_feature_names)
        filtered_test_feature_df = test_features_df[self.filtered_features]
        filtered_test_feature_array = filtered_test_feature_df.to_numpy()
        print(f"Test Filtered feature array Shape : {filtered_test_feature_array.shape}")
        filtered_test_feature_array = np.nan_to_num(filtered_test_feature_array, nan = 0)
        return extracted_features
    


    def get_test_features_array(self):
        test_features_array = extract_features(self.X_test, self.all_feature_names)
        print(test_features_array.shape)
        test_features_df = pd.DataFrame(test_features_array, columns = self.channel_feature_names)
        filtered_test_feature_df = test_features_df[self.filtered_features]
        print(self.channel_feature_names)
        filtered_test_feature_array = filtered_test_feature_df.to_numpy()
        print(f"Test Filtered feature array Shape : {filtered_test_feature_array.shape}")
        filtered_test_feature_array = np.nan_to_num(filtered_test_feature_array, nan = 0)
        return filtered_test_feature_array

    def predict_input(self, input_data):
        if self.classifier is None:
            raise ValueError("No model has not been trained yet to classify new input_data.")
        
        input_features_array = extract_features(input_data, self.all_feature_names)
        input_features_df = pd.DataFrame(input_features_array, columns = self.channel_feature_names)
        filtered_input_feature_df = input_features_df[self.filtered_features]
        filtered_input_feature_array = filtered_input_feature_df.to_numpy()
        filtered_input_feature_array = np.nan_to_num(filtered_input_feature_array, nan = 0)
        predictions = self.classifier.predict(filtered_input_feature_array)

        return predictions
    
    def get_feature_extractor_and_classifier(self):
        return self.feature_extractor, self.classifier

    def display_model_info(self):
        #Needs work
        if self.classifier is None:
            print("The model has not been trained yet.")
            return
        
        print("\nEEG Pipeline Model Information:")
        print("--------------------------------------------------")
        
        print("Selected Features for Training:")
        print(f"  All features: {self.all_feature_names}")
        # print(f"  Filtered features after selection: {self.filtered_features}")
        
        print("\nBest Models and Parameters:")
        best_models_info = [
            (model_name, params)
            for model_name, params in zip(self.best_model_names, self.best_model_params)
        ]
        for model_name, params in best_models_info:
            print(f"  Model: {model_name}")
            print(f"    Parameters: {params}")
        
        print("\nVoting Classifier Information:")
        print(f"  Voting Type: {self.classifier.voting}")
        print("  Base Models:")
        for estimator_name, estimator in self.classifier.named_estimators_.items():
            print(f"    {estimator_name}: {estimator}")

    ##Visualization

    def plot_gaf(self, epoch_to_plot = 0, channel_to_plot = None):
        if channel_to_plot is None:
            channel = 0
        else:
            channel = self.channel_names.index(channel_to_plot)
        eeg_signal = self.data_array[epoch_to_plot][channel]
        gaf = compute_gaf(eeg_signal)
        plot_gaf(gaf)


    #DEEP LEARNING USING IMAGES GENERATED FROM EEG_SIGNALS
    def train_model_gaf(self, training_epochs = 100, training_batch_size = 32):
        max_memory_usage = 0.01  # GB
        image_area = self.data_array.shape[0] ** 2
        n_images = self.data_array.shape[0]

        gaf_scale = ((max_memory_usage * 10**9) / (image_area * n_images))**-1
        print(f"Calculated GAF Scale: {gaf_scale}")

        self.eeg_max = np.max(self.X_train)
        self.eeg_min = np.min(self.X_train)
        print(self.eeg_max)
        print(self.eeg_min)
        gaf_train_images_array = get_gaf_image_array(self.X_train, self.eeg_max, self.eeg_min, gaf_scale=gaf_scale)
        gaf_test_images_array = get_gaf_image_array(self.X_test, self.eeg_max, self.eeg_min, gaf_scale=gaf_scale)
        print(f"Gaf Images Array Shape : {gaf_train_images_array.shape}")
        gaf_classifier, gaf_performance_metrics, val_accuracy, gaf_preds = train_and_evaluate_image_classifier(
            gaf_train_images_array,
            gaf_test_images_array,
            self.y_train,
            num_classes=self.num_classes,
            epochs=training_epochs,
            batch_size=training_batch_size
        )
        self.gaf_classifier = gaf_classifier
        display_image_classifier_performance_metrics(gaf_performance_metrics)
        return gaf_classifier, val_accuracy, gaf_preds




import numpy as np
from sklearn.preprocessing import MinMaxScaler
from scipy import stats
import antropy as ent
import warnings

warnings.filterwarnings("ignore", category=RuntimeWarning)

def mean(x):
    x = np.nan_to_num(x, nan=0)
    return np.mean(x, axis=-1)

def std(x):
    x = np.nan_to_num(x, nan=0)
    return np.std(x, axis=-1)
def ptp(x):
    x = np.nan_to_num(x, nan=0)
    return np.ptp(x, axis=-1)
def var(x):
    x = np.nan_to_num(x, nan=0)
    return np.var(x, axis=-1)
def min(x):
    x = np.nan_to_num(x, nan=0)
    return np.min(x, axis=-1)
def max(x):
    x = np.nan_to_num(x, nan=0)
    return np.max(x, axis=-1)
def argmin(x):
    x = np.nan_to_num(x, nan=0)
    return np.argmin(x, axis=-1)
def argmax(x):
    x = np.nan_to_num(x, nan=0)
    return np.argmax(x, axis=-1)
def rms(x):
    x = np.nan_to_num(x, nan=0)
    return np.sqrt(np.mean(x**2, axis=-1))
def abs_diff_signal(x):
    x = np.nan_to_num(x, nan=0)
    return np.sum(np.abs(np.diff(x, axis=-1)), axis=-1)
def skewness(x):
    x = np.nan_to_num(x, nan=0)
    return stats.skew(x, axis=-1)
def kurtosis(x):
    x = np.nan_to_num(x, nan=0)
    return stats.kurtosis(x, axis=-1)
def signal_energy(x):
    x = np.nan_to_num(x, nan=0)
    return np.sum(x**2, axis=-1)

def hjorth_mobility(x):
    mobilities = []
    for row in x:
        mobility = np.round(ent.hjorth_params(row, axis=-1)[0], 4)
        mobilities.append(mobility)
    return np.array(mobilities)
def hjorth_complexity(x):
    complexities = []
    for row in x:
        complexity = np.round(ent.hjorth_params(row, axis=-1)[1], 4)
        complexities.append(complexity)
    return np.array(complexities)



from setuptools import setup, find_packages

setup(
    name="eeg_ml_pipeline",
    version="0.3",  
    description="A Python package for EEG machine learning analysis, classification, and visualization.",
    author="Yossef Emara",
    author_email="emara004@umn.edu",
    packages=find_packages(),
    install_requires=[
        "numpy",            
        "scikit-learn",
        'pandas>=2.0.0',               
        "scipy",
        "mne",
        "matplotlib",
        "seaborn",
        "antropy",
        "featurewiz",
        "imblearn",
        "tensorflow",
        'dask[dataframe]>=2021.9.1',
        'matplotlib',
        'pyts',
        'librosa',
        'torch',
        'PyWavelets',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import numpy as np
from tqdm import tqdm
from .config import *
import time
import warnings

# Suppress specific UserWarning from scipy.signal
warnings.filterwarnings("ignore", message="nperseg = .* is greater than input length")

def extract_features(data_array, chosen_features, display_progress = True):
    '''
    Purpose: Concatenates desired feature arrays from a data_array based on a chosen_features.
    Parameters:
    - data_array: Array of data where each element contains features to concatenate.
    - chosen_features: List of features to concatenate. Can contain specific features or 'all' to select all features.
    Return Value:
    - features_array: Concatenated feature array.
    '''
    
    def concat_features(x, chosen_features):
        functions = FEATURE_FUNCTIONS_DICT
        
        # selected_functions = [
        #     functions[key](x) for key in functions if key in chosen_features or 'all' in chosen_features
        # ]
        selected_functions = []

        for key in functions:
            if key in chosen_features or 'all' in chosen_features:
                features = functions[key](x)
                features = np.array(features)
                if len(features.shape) == 1:
                    selected_functions.append(features)
                else:
                    for nested_feature in features:
                        selected_functions.append(nested_feature)

        concatenated_features = np.concatenate(selected_functions, axis=-1)
        return concatenated_features

    features = []

    if display_progress:
        print("Extracting features...")
        for d in tqdm(data_array):
            concatenated_features = concat_features(d, chosen_features)
            features.append(concatenated_features)
    else:
        for d in data_array:
            concatenated_features = concat_features(d, chosen_features)
            features.append(concatenated_features)
    
    features_array = np.array(features)
    return features_array

def extract_single_epoch_features(data, chosen_features):
    def concat_features(x, chosen_features):
        functions = FEATURE_FUNCTIONS_DICT
        
        selected_functions = [
            functions[key](x) for key in functions if key in chosen_features or 'all' in chosen_features
        ]
        concatenated_features = np.concatenate(selected_functions, axis=-1)
        return concatenated_features
    
    feature_array = concat_features(data, chosen_features)
    return feature_array


def estimate_feature_extraction_time(data_array):
    '''
    Purpose: Estimates the time taken to extract certain features from a data_array.
    Parameters:
    - data_array: Array of data where each element contains features to concatenate.
    Return Value:
    - elapsed_time_array: An array pf the expected time taken to extract features from
    time, frequency, entropy, and all.
    '''
    num_epochs = data_array.shape[0]
    factor = int(num_epochs/10)

    time_start = time.time()
    extract_features(data_array[:10], TIME_FEATURES_ARRAY, display_progress = False)
    time_end = time.time()
    
    frequency_start = time.time()
    extract_features(data_array[:10], FREQUENCY_FEATURES_ARRAY, display_progress = False)
    frequency_end = time.time()

    entropy_start = time.time()
    extract_features(data_array[:10], FREQUENCY_FEATURES_ARRAY, display_progress = False)
    entropy_end = time.time()

    time_elapsed = (time_end - time_start)*factor
    frequnecy_elapsed = (frequency_end - frequency_start)*factor
    entropy_elapsed = (entropy_end - entropy_start)*factor
    
    all_elpased = time_elapsed + entropy_elapsed + frequnecy_elapsed
    elapsed_time_array = {
        "time" : time_elapsed,
        "frequency" :frequnecy_elapsed,
        "entropy" : entropy_elapsed,
        "all" :all_elpased
        }
    
    return elapsed_time_array




import numpy as np
import matplotlib.pyplot as plt
from pyts.image import GramianAngularField
from .config import *
from scipy.signal import spectrogram, gaussian
from .config import *

def compute_gaf(eeg_signal, eeg_min, eeg_max, scale = 1):
    #Get the min and max using the min, max of the eeg 
    # eeg_min, eeg_max = eeg_signal.min(), eeg_signal.max()
    normalized_signal = 2 * (eeg_signal - eeg_min) / (eeg_max - eeg_min) - 1
    normalized_signal = np.nan_to_num(normalized_signal, nan = 0)
    gaf_transformer = GramianAngularField(image_size=int(len(normalized_signal)//scale), method='difference')
    gaf = gaf_transformer.fit_transform(normalized_signal.reshape(1, -1))[0]
    return gaf

def plot_gaf(gaf):
    plt.figure(figsize=(6, 6))
    plt.imshow(gaf, cmap='rainbow', origin='upper')
    plt.colorbar()
    plt.title("Gramian Angular Field (GAF) of EEG Signal")
    plt.xlabel("Time Index")
    plt.ylabel("Time Index")
    plt.show()

def compute_recurrence_plot(eeg_signal):
    pass

def compute_spectrogram(eeg_signal, fs):

    window_length = SPECTOGRAM_WINDOW_LENGTH
    std_dev = SPECTOGRAM_STD_DEV 
    hop_size = SPECTOGRAM_HOP_SIZE
    mfft = SPECTOGRAM_MFFT

    window = gaussian(window_length, std=std_dev, sym=True)

    f, t, Sxx = spectrogram(eeg_signal, fs, window=window, noverlap=window_length-hop_size, nfft=mfft)

    return Sxx[:int(len(Sxx)/4)], f, t

def plot_spectrogram(Sxx, f, t):
    plt.figure(figsize=(10, 6))
    plt.imshow(Sxx, origin='lower', aspect='auto', cmap='viridis', extent=[t.min(), t.max(), f.min(), SPECTROGRAM_MAX_FREQUENCY])
    plt.colorbar(label="Frequency")
    plt.title("EEG Signal Spectrogram")
    plt.xlabel("Time [s]")
    plt.ylabel("Frequency [Hz]")
    plt.tight_layout()
    plt.show()
